-- Columnar Storage

USE AdventureworksDW2014;
GO
SET NOCOUNT ON;
GO

IF OBJECT_ID(N'dbo.InternetSales', N'U') IS NOT NULL
   DROP TABLE dbo.InternetSales;
GO
-- Fact table
CREATE TABLE dbo.InternetSales
(
 InternetSalesKey INT      NOT NULL IDENTITY(1,1),
 CustomerKey      INT      NOT NULL,
 ProductKey       INT      NOT NULL,
 DateKey          INT      NOT NULL,
 OrderQuantity    SMALLINT NOT NULL DEFAULT 0,
 SalesAmount      MONEY    NOT NULL DEFAULT 0,
 UnitPrice        MONEY    NOT NULL DEFAULT 0,
 DiscountAmount   FLOAT    NOT NULL DEFAULT 0,
);
GO
-- UCLI
CREATE UNIQUE CLUSTERED INDEX XCLI_InternetSales
 ON dbo.InternetSales(InternetSalesKey);
GO

/* With DROP_EXISTING
-- UCLI
CREATE UNIQUE CLUSTERED INDEX XCLI_InternetSales
 ON dbo.InternetSales(InternetSalesKey)
 WITH DROP_EXISTING;
GO
*/

-- Insert data into InternetSales
INSERT INTO dbo.InternetSales
(CustomerKey, 
 ProductKey, DateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount)
SELECT 
 CustomerKey, 
 ProductKey, OrderDateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount
FROM dbo.FactInternetSales;
GO 30

-- Check the number of rows in the dbo.FactInternetSales table
SELECT COUNT(*), COUNT(*) / 30
FROM dbo.InternetSales;
GO

-- Check the space used by the InternetSales table
EXEC sys.sp_spaceused N'dbo.InternetSales',
 @updateusage = N'TRUE';
GO

-- Show execution plan
-- Set statistics on
SET STATISTICS IO ON;
GO

-- Simple query
SELECT I.ProductKey,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
WHERE I.ProductKey < 300
GROUP BY I.ProductKey
ORDER BY I.ProductKey;
GO
-- Full scan - 11512
-- Only row mode operators

-- Complex query
SELECT G.EnglishCountryRegionName AS CountryRegion,
 PC.EnglishProductCategoryName AS Product,
 D.CalendarYear,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
 INNER JOIN dbo.DimCustomer AS C
  ON I.CustomerKey = C.CustomerKey
 INNER JOIN dbo.DimGeography AS G
  ON C.GeographyKey = G.GeographyKey
 INNER JOIN dbo.DimProduct AS P
  ON I.ProductKey = p.ProductKey
 INNER JOIN dbo.DimProductSubcategory AS PS
  ON P.ProductSubcategoryKey = PS.ProductSubcategoryKey
 INNER JOIN DimProductCategory AS PC
  ON PS.ProductCategoryKey = PC.ProductCategoryKey
 INNER JOIN dbo.DimDate AS d
  ON I.DateKey = D.DateKey
GROUP BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
ORDER BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear;
GO
-- Full scan - 11512
-- Only row mode operators

-- Point query
SELECT *
FROM dbo.InternetSales
WHERE CustomerKey = 21768;
-- Full scan - 11512

SET STATISTICS IO OFF;
GO

-- NCCI
-- Create
CREATE NONCLUSTERED COLUMNSTORE INDEX NCCI_InternetSales
  ON dbo.InternetSales
  (InternetSalesKey, 
   CustomerKey,
   ProductKey, DateKey,
   OrderQuantity, SalesAmount,
   UnitPrice, DiscountAmount);
GO

-- Re-check the space used by the InternetSales table
EXEC sys.sp_spaceused N'dbo.InternetSales',
 @updateusage = N'TRUE';
GO

SET STATISTICS IO ON;
GO

-- Simple query, no NCCI
SELECT I.ProductKey,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
WHERE I.ProductKey < 300
GROUP BY I.ProductKey
ORDER BY I.ProductKey
OPTION (ignore_nonclustered_columnstore_index);
GO
-- Full scan - 11512

-- Complex query, no NCCI
SELECT G.EnglishCountryRegionName AS CountryRegion,
 PC.EnglishProductCategoryName AS Product,
 D.CalendarYear,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
 INNER JOIN dbo.DimCustomer AS C
  ON I.CustomerKey = C.CustomerKey
 INNER JOIN dbo.DimGeography AS G
  ON C.GeographyKey = G.GeographyKey
 INNER JOIN dbo.DimProduct AS P
  ON I.ProductKey = p.ProductKey
 INNER JOIN dbo.DimProductSubcategory AS PS
  ON P.ProductSubcategoryKey = PS.ProductSubcategoryKey
 INNER JOIN DimProductCategory AS PC
  ON PS.ProductCategoryKey = PC.ProductCategoryKey
 INNER JOIN dbo.DimDate AS d
  ON I.DateKey = D.DateKey
GROUP BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
ORDER BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
OPTION (ignore_nonclustered_columnstore_index);
GO
-- Full scan - 11512

-- Point query, no NCCI
SELECT *
FROM dbo.InternetSales
WHERE CustomerKey = 21768
OPTION (ignore_nonclustered_columnstore_index);
-- Full scan - 11512

-- Simple query, NCCI
SELECT I.ProductKey,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
WHERE I.ProductKey < 300
GROUP BY I.ProductKey
ORDER BY I.ProductKey;
GO
-- NCCI scan - 1248

-- Complex query, NCCI
SELECT G.EnglishCountryRegionName AS CountryRegion,
 PC.EnglishProductCategoryName AS Product,
 D.CalendarYear,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
 INNER JOIN dbo.DimCustomer AS C
  ON I.CustomerKey = C.CustomerKey
 INNER JOIN dbo.DimGeography AS G
  ON C.GeographyKey = G.GeographyKey
 INNER JOIN dbo.DimProduct AS P
  ON I.ProductKey = p.ProductKey
 INNER JOIN dbo.DimProductSubcategory AS PS
  ON P.ProductSubcategoryKey = PS.ProductSubcategoryKey
 INNER JOIN DimProductCategory AS PC
  ON PS.ProductCategoryKey = PC.ProductCategoryKey
 INNER JOIN dbo.DimDate AS d
  ON I.DateKey = D.DateKey
GROUP BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
ORDER BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear;
GO
-- NCCI scan - 2778

-- Point query, NCCI
SELECT *
FROM dbo.InternetSales
WHERE CustomerKey = 21768;
-- NCCI scan - 2813

SET STATISTICS IO OFF;
GO

-- Drop the NCCI
DROP INDEX NCCI_InternetSales
  ON dbo.InternetSales;
GO

-- Create a CCI
CREATE CLUSTERED COLUMNSTORE INDEX XCLI_InternetSales
  ON dbo.InternetSales
  WITH DROP_EXISTING;
GO

-- Re-check the space used by the InternetSales table
EXEC sys.sp_spaceused N'dbo.InternetSales',
 @updateusage = N'TRUE';
GO

SET STATISTICS IO ON;
GO

-- Simple query, CCI
SELECT I.ProductKey,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
WHERE I.ProductKey < 300
GROUP BY I.ProductKey
ORDER BY I.ProductKey;
GO
-- CCI segment scan - 26

-- Complex query, CCI
SELECT G.EnglishCountryRegionName AS CountryRegion,
 PC.EnglishProductCategoryName AS Product,
 D.CalendarYear,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
 INNER JOIN dbo.DimCustomer AS C
  ON I.CustomerKey = C.CustomerKey
 INNER JOIN dbo.DimGeography AS G
  ON C.GeographyKey = G.GeographyKey
 INNER JOIN dbo.DimProduct AS P
  ON I.ProductKey = p.ProductKey
 INNER JOIN dbo.DimProductSubcategory AS PS
  ON P.ProductSubcategoryKey = PS.ProductSubcategoryKey
 INNER JOIN DimProductCategory AS PC
  ON PS.ProductCategoryKey = PC.ProductCategoryKey
 INNER JOIN dbo.DimDate AS d
  ON I.DateKey = D.DateKey
GROUP BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
ORDER BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear;
GO
-- CCI scan - 1556

-- Point query, CCI
SELECT *
FROM dbo.InternetSales
WHERE CustomerKey = 21768;
-- CCI scan - 2813

SET STATISTICS IO OFF;
GO

-- Archive compression
ALTER INDEX XCLI_InternetSales
 ON dbo.InternetSales
 REBUILD WITH (DATA_COMPRESSION = COLUMNSTORE_ARCHIVE);
GO

-- Re-check the space used by the InternetSales table
EXEC sys.sp_spaceused N'dbo.InternetSales',
 @updateusage = N'TRUE';
GO

SET STATISTICS IO ON;
GO

-- Simple query, CCI + ARCH
SELECT I.ProductKey,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
WHERE I.ProductKey < 300
GROUP BY I.ProductKey
ORDER BY I.ProductKey;
GO
-- CCI segment scan - 12

-- Complex query, CCI + ARCH
SELECT G.EnglishCountryRegionName AS CountryRegion,
 PC.EnglishProductCategoryName AS Product,
 D.CalendarYear,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
 INNER JOIN dbo.DimCustomer AS C
  ON I.CustomerKey = C.CustomerKey
 INNER JOIN dbo.DimGeography AS G
  ON C.GeographyKey = G.GeographyKey
 INNER JOIN dbo.DimProduct AS P
  ON I.ProductKey = p.ProductKey
 INNER JOIN dbo.DimProductSubcategory AS PS
  ON P.ProductSubcategoryKey = PS.ProductSubcategoryKey
 INNER JOIN DimProductCategory AS PC
  ON PS.ProductCategoryKey = PC.ProductCategoryKey
 INNER JOIN dbo.DimDate AS d
  ON I.DateKey = D.DateKey
GROUP BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
ORDER BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear;
GO
-- CCI scan - 225

-- Point query, CCI + ARCH
SELECT *
FROM dbo.InternetSales
WHERE CustomerKey = 21768;
-- CCI scan - 1445

SET STATISTICS IO OFF;
GO

-- NCI
CREATE NONCLUSTERED INDEX XNCLI_InternetSales
 ON dbo.InternetSales(CustomerKey);
GO

-- Re-check the space used by the InternetSales table
EXEC sys.sp_spaceused N'dbo.InternetSales',
 @updateusage = N'TRUE';
GO

SET STATISTICS IO ON;
GO

-- Simple query, CCI + ARCH + NCI
SELECT I.ProductKey,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
WHERE I.ProductKey < 300
GROUP BY I.ProductKey
ORDER BY I.ProductKey;
GO
-- CCI segment scan - 12

-- Complex query, CCI + ARCH + NCI
SELECT G.EnglishCountryRegionName AS CountryRegion,
 PC.EnglishProductCategoryName AS Product,
 D.CalendarYear,
 SUM(I.SalesAmount) AS Sales
FROM dbo.InternetSales AS I
 INNER JOIN dbo.DimCustomer AS C
  ON I.CustomerKey = C.CustomerKey
 INNER JOIN dbo.DimGeography AS G
  ON C.GeographyKey = G.GeographyKey
 INNER JOIN dbo.DimProduct AS P
  ON I.ProductKey = p.ProductKey
 INNER JOIN dbo.DimProductSubcategory AS PS
  ON P.ProductSubcategoryKey = PS.ProductSubcategoryKey
 INNER JOIN DimProductCategory AS PC
  ON PS.ProductCategoryKey = PC.ProductCategoryKey
 INNER JOIN dbo.DimDate AS d
  ON I.DateKey = D.DateKey
GROUP BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear
ORDER BY G.EnglishCountryRegionName,
 PC.EnglishProductCategoryName, D.CalendarYear;
GO
-- CCI scan - 225

-- Point query, CCI + ARCH + NCI
SELECT *
FROM dbo.InternetSales
WHERE CustomerKey = 21768;
-- NCI seek - 3 & CI key lookup

SET STATISTICS IO OFF;
GO

-- Drop the NCI index
DROP  INDEX dbo.InternetSales.XNCLI_InternetSales;
GO

-- Check the catalog views
-- Check the segments
SELECT *
FROM sys.column_store_dictionaries;
SELECT *
FROM sys.column_store_segments;
SELECT OBJECT_NAME(object_id), *
FROM sys.column_store_row_groups; 
GO
-- All row groups compressed

-- Insert data into InternetSales
-- 60,398 rows
INSERT INTO dbo.InternetSales
(CustomerKey, 
 ProductKey, DateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount)
SELECT 
 CustomerKey, 
 ProductKey, OrderDateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount
FROM dbo.FactInternetSales;
GO

-- Row groups
SELECT *
FROM sys.column_store_row_groups; 
GO
-- 1 new row group, open, 60,398 rows

-- Insert data into InternetSales
-- 966,368 rows
INSERT INTO dbo.InternetSales
(CustomerKey, 
 ProductKey, DateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount)
SELECT 
 CustomerKey, 
 ProductKey, OrderDateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount
FROM dbo.FactInternetSales;
GO 16

-- Row groups
SELECT *
FROM sys.column_store_row_groups; 
GO
-- 1 new row group, open, 1,026,766 rows

-- Insert data into InternetSales
-- 60,398 rows
INSERT INTO dbo.InternetSales
(CustomerKey, 
 ProductKey, DateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount)
SELECT 
 CustomerKey, 
 ProductKey, OrderDateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount
FROM dbo.FactInternetSales;
GO

-- Row groups
SELECT *
FROM sys.column_store_row_groups; 
GO
-- 1 new new row group, open,      38,588 rows
-- 1 old new row group, closed, 1,048,576 rows

-- Kick the tuple mover
-- ALTER TABLE dbo.InternetSales REBUILD;
-- The above closes and compresses also the small row group
-- Use rather ALTER INDEX
ALTER INDEX XCLI_InternetSales
 ON dbo.InternetSales REORGANIZE;
GO

-- Row groups
SELECT *
FROM sys.column_store_row_groups; 
GO
-- 1 new new row group, open,          38,588 rows
-- 1 old new row group, tombstone,  1,048,576 rows

-- Rebuild with columnstore compression
ALTER INDEX XCLI_InternetSales
 ON dbo.InternetSales
 REBUILD WITH (DATA_COMPRESSION = COLUMNSTORE);
GO

-- Row groups
SELECT *
FROM sys.column_store_row_groups; 
GO
-- All row groups compressed

-- Insert data into InternetSales
-- 120,796 rows
INSERT INTO dbo.InternetSales
(CustomerKey, 
 ProductKey, DateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount)
SELECT 
 CustomerKey, 
 ProductKey, OrderDateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount
FROM dbo.FactInternetSales
UNION ALL
SELECT 
 CustomerKey, 
 ProductKey, OrderDateKey,
 OrderQuantity, SalesAmount,
 UnitPrice, DiscountAmount
FROM dbo.FactInternetSales;
GO

-- Row groups
SELECT *
FROM sys.column_store_row_groups; 
GO
-- Bulk insert over 100,000 rows
-- A new compressed row group created

-- Clean up
IF OBJECT_ID(N'dbo.InternetSales', N'U') IS NOT NULL
   DROP TABLE dbo.InternetSales;
GO

